package PureMessage::WebUI;
use strict;
use warnings;
use base 'CGI::Application';
use CGI::Application::Plugin::TT;
use CGI::Application::Plugin::Authentication;
use CGI::Application::Plugin::Session;
use PureMessage::Version qw($PREFIX);
use Apache::Constants qw(OK DECLINED REDIRECT FORBIDDEN SERVER_ERROR);
use Net::Domain ();
use POSIX qw(strftime);
use PureMessage::User qw/login_validator $LOGIN_ERROR/;
use PureMessage::WebUI::Log qw/activity_log/;
use PureMessage::WebUI::I18N;
use PureMessage::WebUI::Controller::Search  qw(handle_search load_search_reasons);
use PureMessage::WebUI::Controller::Message qw(message_details print_message_attachment
                                               forward_message show_original);
use PureMessage::WebUI::Controller::Action  qw(handle_action save_mbox);
use PureMessage::WebUI::Controller::Report  qw(handle_report);
use PureMessage::WebUI::Controller::Configuration qw(handle_config_tab);
use PureMessage::WebUI::Controller::Password qw(handle_change_password);
use PureMessage::WebUI::Controller::LogSearch qw(handle_log_search handle_logsearch_preview);
use PureMessage::WebUI::Utils qw($DEFAULT_DATE utf8_truncate_filter);
use PureMessage::WebUI::RunModeException;
use PureMessage::WebUI::ExpireSessions qw/$WEBUI_CACHE_DIR/;
use PureMessage::_sys;
use File::Path qw(mkpath);
use ActiveState::Handy qw(cat);
use JSON::Syck;
use Template::Stash;

our %login_config = (
    authen_config => {
        DRIVER => ['Generic', \&login_validator],
        STORE  => 'Session',
        POST_LOGIN_RUNMODE => 'post_login',
        LOGIN_RUNMODE => 'login',
    },
    # Everything but login and logout is protected
    protected_runmodes => qr/^(?!login|logout)/,
);

my $LOGSEARCH_ACTIONS = [qw(Bounced Copy Deferred Deliver Discard Quarantine Reject Route)];

# Initialize the CGI::App object
# This is called right before setup()
sub cgiapp_init {
    my $self = shift;

    # Global Template Toolkit config
    my $tmpl_dir = "$PREFIX/var/www/group-tmpl";
    $self->tt_config( TEMPLATE_OPTIONS => {
        INCLUDE_PATH => $tmpl_dir,
        STASH => Template::Stash->new,
        FILTERS => {
            'utf8_truncate' => [\&utf8_truncate_filter, 1],
        },
    });

    # Authentication and Session setup
    $self->authen->config($login_config{authen_config});
    $self->authen->protected_runmodes($login_config{protected_runmodes});
    $self->session_config(
        CGI_SESSION_OPTIONS => [ 
            "driver:File", 
            $self->query, 
            { Directory => "$WEBUI_CACHE_DIR/sessions" } 
        ]);
}

sub setup {
    my $self = shift;
    $self->start_mode('search');
    $self->error_mode('show_error');
    $self->mode_param(path_info => 1);
    $self->run_modes(
        search  =>  'show_search',
        logsearch => 'show_logsearch',
        logview => 'show_logsearch_preview',
        results =>  'show_results',
        message =>  'show_message', # message detail popups
        forward =>  'show_forward', # message forward popup
        action  =>  'perform_action', # ajax actions
        reports =>  'show_report',
        'report-img' => 'show_report_asset',
        config  =>  'show_config',
        login   =>  'show_login',
        logout  =>  'show_logout',
        post_login => 'post_login',
        prohibited => 'show_prohibited',
        save => 'save_messages',
        original => 'fetch_original',
        password => 'show_change_password',

        AUTOLOAD => 'show_error',
    );
    
    # Uses the pmx-config language setting (default english)
    $self->{lh} = PureMessage::WebUI::I18N->get_handle_via_config();
}

sub http_return_code {
    my ($self, $code) = @_;
    if($code){
        $self->{http_return_code} = $code;
    }
    return $self->{http_return_code} || OK;
}

# This is called after setup, but right before the runmode runs
sub cgiapp_prerun {
    my ($self, $run_mode) = @_;

    # Setup for all runmodes
    my $username = $self->authen->username || $login_config{default_username};
    $self->{context}{username} = $username;
    my $user;
    my $lh = $self->{lh};
    my $loc = sub { $lh->maketext(@_) };
    my $img = sub { $lh->make_image_path(@_) };
    if ($username) { 
        eval {
            $user = PureMessage::User->load($self->{context}{username});
        };
        warn $@ if $@;
        if ($@ and $@ =~ /DBI connect/) {
            $self->{context}{login_msg} = $loc->(db_unavailable_msg());
        }
        $self->prerun_mode('login') if $@;
        $self->{context}{user} = $user;
    }

    my $server_time = strftime(q{%H:%M}, localtime(time()));
    %{ $self->{context} } = ( %{ $self->{context} }, 
      server => Net::Domain::hostname(),
      'time' => $server_time,
      search_results => [],
      statusbar => 'inactive',
      loc => $loc,
      img => $img,
      
      # These are used by the status_bar.tmpl and msg_status_bar.tmpl
      # templates.  We cannot use the loc function in these templates
      # because they are also parsed by Jemplate in Ajax requests, and
      # therefore cannot use the Perl loc function directly.
      lang => {
        confirm_deleteall => $loc->('Warning: This will delete all messages in your current query.  Continue?'),
        btn_delete_all => $loc->('Delete All'),
        btn_delete => $loc->('Delete'),
        btn_save => $loc->('Save'),
        btn_forward => $loc->('Forward'),
        btn_approve => $loc->('Approve'),
        btn_close => $loc->('Close'),
        note => $loc->('Note'),
      },
    );

    # ajax actions don't need extra permissions and
    # unit tests call cgiapp_prerun('')
    return if $run_mode eq 'action' or $run_mode eq '';
    return unless $user;

    # find tab permissions
    $self->{context}{perm}{pages} = $user->any_group_can(
        qw(configuration help reports quarantine quarantine.preview
           change_password logsearch logsearch.preview));

    # The 'search' tab is now a combination of things, so we need a new
    # permission for it
    $self->{context}{perm}{pages}{search_tab} = (
        $self->{context}{perm}{pages}{quarantine}
        || $self->{context}{perm}{pages}{logsearch}
    );

    # don't show the config tab if no groups exist (except supergroup)
    if ($user->super_user and $user->groups == 1) {
        $self->{context}{perm}{pages}{configuration} = 0;
    }

    # Check that user has access to the page they're going to
    my %run_mode_to_perm = (search => 'quarantine',
                            logsearch => 'logsearch',
                            results => 'quarantine',
                            message => 'quarantine.preview',
                            logview => 'logsearch.preview',
                            reports => 'reports', 
                            config => 'configuration',
                           );
    my $run_mode_perm = $run_mode_to_perm{$run_mode};
    return unless $run_mode_perm;
    unless ($self->{context}{perm}{pages}{$run_mode_perm}) {
        $self->log("No access to run mode: $run_mode");
        $self->{context}{msg_prohibited}++ if $run_mode eq 'message';
        $self->prerun_mode('prohibited');
    }
}

# runmode that will redirect to appropriate runmode
# after a user has logged in
sub post_login {
    my ($self) = @_;
    my $user = $self->{context}{user};
    my $perms = $user->any_group_can(qw(quarantine logsearch configuration reports));
    my $pages = [[quarantine => \&show_search ],
                 [logsearch => \&show_logsearch ],
                 [configuration => \&show_config ],
                 [reports => \&show_report ]];

    for my $page (@$pages){
        if ($perms->{$page->[0]}) {
            my $code = $page->[1];
            return $self->$code();
        }
    }
    return $self->show_prohibited;
}

sub show_login {
    my ($self) = @_;

    my $context = $self->{context};
    if ($self->authen->login_attempts) {
        $context->{login_msg} = $context->{loc}->("Invalid username or password");
    }
    my $q = $self->query;
    my $user = $q->param('authen_username');
    my $pass = $q->param('authen_username');
    if ((defined($user) and !$user) or (defined($pass) and !$pass)) {
        $context->{login_msg} = 
            $context->{loc}->("Please enter a username and password");
    }

    # $LOGIN_ERROR is set by the login_validator in PM::User
    if ($LOGIN_ERROR and $LOGIN_ERROR =~ /DBI connect/) {
        $context->{login_msg} = $context->{loc}->(db_unavailable_msg());
    }
    
    return $self->tt_process(q{login.tmpl}, $context);
}

sub show_logout {
    my ($self) = @_;
    if ($self->authen->is_authenticated()) {
        $self->log('User logged out');
        $self->authen->logout();
        $self->session->delete;
    }
    my $context = $self->{context};
    $context->{login_msg} = $context->{loc}->('Logged out');
    
    return $self->tt_process(q{login.tmpl}, $context);
}

sub show_config {
    my ($self) = @_;
    handle_config_tab($self);
    return $self->tt_process($self->{context}{tmpl}, $self->{context});
}

sub show_search {
    my ($self) = @_;
    my $context = $self->{context};
    $context->{tab} = q{search};
    $context->{page} = q{default};
    $context->{searchtype} = q{quarantine};
    $context->{actions} = $LOGSEARCH_ACTIONS;
    $context->{reasons} = load_search_reasons($self);
    $context->{logreasons} = _gain_log_reasons();
    $context->{search}{params}{ldateFrom} = $DEFAULT_DATE;
    $context->{search}{params}{ldateTo} = $DEFAULT_DATE;
    $context->{search}{params}{qdateFrom} = $DEFAULT_DATE;
    $context->{search}{params}{qdateTo} = $DEFAULT_DATE;
    return $self->tt_process(q{search.tmpl}, $context);
}

sub _gain_log_reasons {
    use PureMessage::List;
    my $list = PureMessage::List->new('log-reasons');
    my $ld = $list->data();
    my $reason_list = [];
    $ld->enumerate_items(sub { my $item = shift; push @$reason_list, $item; });
    return $reason_list;
}

sub show_logsearch {
    my ($self) = @_;
    my $context = $self->{context};
    $context->{tab} = q{search};
    $context->{page} = q{logsearch};
    $context->{actions} = $LOGSEARCH_ACTIONS;
    $context->{reasons} = load_search_reasons($self);
    $context->{logreasons} = _gain_log_reasons();
    $context->{search}{params}{ldateFrom} = $DEFAULT_DATE;
    $context->{search}{params}{ldateTo} = $DEFAULT_DATE;
    $context->{search}{params}{qdateFrom} = $DEFAULT_DATE;
    $context->{search}{params}{qdateTo} = $DEFAULT_DATE;

    # XXX - 
    #   Need to load reasons
    #   Need to load actions
    if( $self->query->param('logsearch') ){
        handle_log_search($self);
    }
    return $self->tt_process(q{logsearch.tmpl}, $context);
}

sub show_logsearch_preview {
    my ($self) = @_;
    my $context = $self->{context};
    $context->{can_view_rawlogs} = 
            $context->{user}->any_group_can('logsearch.preview.rawlogs');

    handle_logsearch_preview($self);
    return $self->tt_process(q{logpreview.tmpl}, $context);
}

sub show_results {
    my ($self) = @_;
    my $context = $self->{context};
    $context->{actions} = $LOGSEARCH_ACTIONS;
    $context->{reasons} = load_search_reasons($self);
    $context->{search}{params}{ldateFrom} = $DEFAULT_DATE;
    $context->{search}{params}{ldateTo} = $DEFAULT_DATE;
    $context->{search}{params}{qdateFrom} = $DEFAULT_DATE;
    $context->{search}{params}{qdateTo} = $DEFAULT_DATE;
    $context->{logreasons} = _gain_log_reasons();
    $self->load_action_perms;

    handle_search($self);
    return $self->tt_process('search.tmpl', $context);
}

sub show_report {
    my ($self) = @_;
    return handle_report($self, 'table');
}

sub show_report_asset {
    my ($self) = @_;
    return handle_report($self, 'asset');
}

sub show_message {
    my ($self) = @_;

    my ($page, $msg_id, $part_id);
    eval {
        $self->load_action_perms;
        my $path_info = $self->query->path_info();
        die qq{Can't parse path_info: $path_info}
            unless $path_info =~ m#^/message/(\w+)/([\d-]+\d)(?:/([\d\.]+\d))?$#;
        ($page, $msg_id, $part_id) = ($1, $2, $3);
        if ($part_id) {
            # this function will tell CGI-App not to send headers,
            # will print it's own headers and then send out the 
            # attachment in chunks to the browser.
            print_message_attachment($self, $msg_id, $part_id);
        }
        else {
            message_details($self, $page, $msg_id);
        }
    };
    die $@ if ref($@);
    die PureMessage::WebUI::MessagePreviewException->new(error_msg => $@) if $@;

    return '' if $part_id;
    return $self->tt_process(qq{message/$page.tmpl}, $self->{context});
}

sub show_forward {
    my ($self) = @_;

    eval {
        $self->load_action_perms;
        forward_message($self);
    };
    die $@ if ref($@);
    die PureMessage::WebUI::MessagePreviewException->new(error_msg => $@) if $@;

    return $self->tt_process(qq{message/forward.tmpl}, $self->{context});
}

sub save_messages {
    my ($self) = @_;
    my $mbox = $self->query->param('mbox');
    eval {
        save_mbox($self, $mbox);
    };
    if($mbox){
        unlink $mbox or die qq{Could not remove $mbox};
    }
    die $@ if $@;

    return '';
}

sub fetch_original {
    my ($self) = @_;
    my $msg = $self->query->param('msg');
    my $source = undef;
    eval { $source = show_original($self, $msg); };
    die $@ if $@;
    $self->header_props('-type' => 'text/plain');
    return $source;
}

sub show_change_password {
    my ($self) = @_;

    eval { handle_change_password($self) };
    die $@ if ref($@);
    die PureMessage::WebUI::MessagePreviewException->new(error_msg => $@) if $@;

    return $self->tt_process(qq{change_password.tmpl}, $self->{context});
}

# load the permissions for the message action buttons
sub load_action_perms {
    my $self = shift;
    my $context = $self->{context};
    my $user = $context->{user};
    if ($user->any_group_can('quarantine.actions')) {
        my @actions = qw(delete approve forward save);
        my $perms = $user->any_group_can( map { "quarantine.actions.$_" } @actions );
        $context->{perm}{actions} = { map { ( $_ => $perms->{"quarantine.actions.$_"} ) }
                                      @actions };
    }
}

# set msg to true for message popup templates
sub show_prohibited {
    my ($self, $msg) = @_;
    $self->http_return_code(FORBIDDEN);
    my $template = 'errors/forbidden.tmpl';
    $template = 'message/forbidden.tmpl' 
        if $msg or $self->{context}{msg_prohibited};
    return $self->tt_process($template, $self->{context});
}

# Action Ajax handler
sub perform_action {
    my $self = shift;

    my $context;
    eval {
        $self->load_action_perms;
        $context = handle_action($self);
        $context->{perm}{actions} = $self->{context}{perm}{actions};
        $context->{lang} = $self->{context}{lang};
    };
    my $err = $@;
    if ($err) {
        warn $err;
        $self->log($err);
        $context->{statusbar} = 'bad';
        $context->{note} = $err;
    }

    return JSON::Syck::Dump($context);
}

sub log {
    my ($self, @msgs) = @_;
    activity_log($self->{context}{username}, @msgs);
}

sub show_error {
    my ($self, $err) = @_;
    if (UNIVERSAL::can($err, 'error_message')) {
        $self->log("Error: " . $err->error_message) if $err->error_message;
        return $self->show_prohibited($err->{message_preview}) if $err->prohibited;
        if (my $url = $err->redirect) {
            $self->header_type('redirect');
            $self->header_props(-url=>$url);
            return "Redirecting to $url\n";
        }
        if ($err->{message_preview}) {
            my $context = $self->{context};
            $context->{error} = $err->error_message 
                  . "(The message may have been deleted or there "
                  . "could be a network issue when fetching the message.)";
            return $self->tt_process('message/error.tmpl', $context);
        }
        # default to just an error message
        $err = $err->error_message;
    }
    $self->log("Error: $err");
    return $self->tt_process('errors/general.tmpl', { 
            error_msg => $err, loc => $self->{context}{loc} 
    });
}

# we apparently need this to make sure session state
# is written to disk
sub teardown {
    my $self = shift;
    $self->session->flush();
}

sub db_unavailable_msg {
    return "This service is unavailable, please contact your "
           . "administrator for assistance.";
}

##################
# Package Methods
# ################
 
# Mod_perl handler
sub handler {
    my $r = shift;

    my $webapp = PureMessage::WebUI->new();
    $webapp->header_props('-type' => 'text/html; charset=UTF-8');
    eval {
        $webapp->run();
    };
    if ($@) {
        # shouldn't happen anymore - errors trapped in show_error runmode
        warn $@;
        $r->headers_out->set(Location => '/groups');
        $r->status(REDIRECT);
        $r->send_http_header;
    }

    my $rc = $webapp->http_return_code;
    $r->status($rc) if $rc != OK;
    return OK;
}

1;
