package PureMessage::WebUI::Controller::Action;
use strict;
use warnings;
use base 'Exporter';
our @EXPORT_OK = qw(handle_action save_mbox);
use PureMessage::Version qw/$PREFIX/;
use PureMessage::QMan qw(approve trash full_message_source);
use PureMessage::Logger;
use PureMessage::WebUI::Utils qw(get_message_status mset_query
                                 calc_message_status
                                 message_status_from_id);
use PureMessage::Group::CDBMap;
use PureMessage::WebUI::Controller::Search qw(search_reasons_per_group);
use PureMessage::WebUI::ExpireSessions qw/$WEBUI_CACHE_DIR/;

use File::Temp qw(tempfile);
use PureMessage::Mbox qw(mbox_append);
use POSIX qw(strftime);

my $q_logger = PureMessage::Logger->new(ident => "GroupsUI");

sub handle_action {
    my ($app) = @_;

    # determine which action to call
    my $path_info = $app->query->path_info;
    die "Can't parse path_info: $path_info"
        unless $path_info =~ m#^/action/(\w+)#;
    my $action = $1;
    my %valid_action = map { $_ => 1 } 
                       qw(approve delete delete_all forward save report processing);
    die "$action isn't a valid message action!" unless $valid_action{$action};

    # handle actions
    my %context;
    if ($action eq 'processing') {
        unless ($path_info =~ m#^/\w+/processing/(\w+)#) {
            die "Unknown path: $path_info";
        }
        my $user_action = $1;
        %context = _handle_processing($app, $user_action);
    }
    elsif ($action eq 'approve') {
        %context = _handle_approve($app);
    }
    elsif ($action eq 'delete') {
        %context = _handle_delete($app);
    }
    elsif ($action eq 'delete_all') {
        %context = _handle_delete_all($app);
    }
    elsif ($action eq 'save') {
        %context = _handle_save($app);
    }
    $app->header_props(-type => 'text/plain; charset=UTF-8');        
    
    return \%context;
}

sub save_mbox {
    my ($app, $mbox) = @_;
    die "Cannot fetch mbox path for action SAVE" unless $mbox;

    my $size = (stat $mbox)[7];

    open(my $fh, $mbox) or die "Cannot open mbox '$mbox' for read: $!";
    $app->header_type('none');
    print $app->query->header(
        '-type' => 'text/plain',
        '-Content-Disposition' => 'attachment; filename=messages-' . strftime("%Y%m%d-%H%M%S", localtime(time)) . '.mbox',
        '-Content-Length' => $size,
    );
    my $now = 0;
    my $buffer = undef;
    while(read($fh, $buffer, 10240)) {
        unless(print $buffer) {
            $app->log("Connection lost, aborting download of mbox: $mbox");
            die "Connection lost for downloading mbox: $mbox";
        }
    }
    close $fh or die "Cannot close mbox '$mbox' after read";
}

sub _handle_processing {
    my ($app, $user_action) = @_;
    my $loc = $app->{context}{loc} || die "Can't find loc!";
    my %context = (statusbar => 'warning');

    if ($user_action eq 'delete_all') {
        my $where = $app->query->param('where');
        if ($where) {
            $context{note} = $loc->('Deleting all messages in current query');
        }
        else {
            $context{note} = $loc->('Please perform a search before deleting all messages');
        }
        return %context;
    }

    my $num = $app->query->param('num');
    if ($num) {
        my %verb = (
            delete  => 'Deleting',
            approve => 'Approving',
            save    => 'Saving',
            forward => 'Forwarding',
        );
        $context{note} = $loc->("$verb{$user_action} [quant,_1,message,messages]", $num);
    }
    else {
        $context{note} = $loc->('Please select a message');
    }    

    return %context;
}

sub _handle_approve {
    my ($app) = @_;
    my $to = $app->query->param('to');
    return $to ? _per_recipient_approve($app, $to) 
               : _handle_approve_delete($app, 'approve');
}

sub _handle_delete {
    my ($app) = @_;
    return _handle_approve_delete($app, 'delete');
}

sub _handle_save {
    my ($app) = @_;
    my $msg_ids = $app->query->param('msgs');
    return unless $msg_ids;

    my $quarantine = PureMessage::Quarantine->new;
    my @msg_ids = split ',', $msg_ids;

    my $total = @msg_ids;
    my $now = 0;

    my $loc = $app->{context}{loc};
    my %context = (
        message => { msg_id => $msg_ids[0] }, # used by msg status bar
    );

    my $mbox_tmp_dir = "$WEBUI_CACHE_DIR/tmp";
    my (undef, $fn) = tempfile(DIR => $mbox_tmp_dir);
    unless($fn) {
        $app->log("Cannot create temporary file for action SAVE");
        return (
            %context,
            statusbar => 'bad', 
            note => $loc->('Saved 0 / [quant,_1,message,messages]', $total),
        );
    }

    for my $id (@msg_ids) {
        my $msg = $quarantine->message($id);
        unless($msg) {
            $app->log("Cannot create message $id");
            next;
        }
        my $ok = _have_permission($app, 'save', $msg);
        unless($ok) {
            $app->log("No permission to save $id");
            next;
        }
        eval {
            my $source = full_message_source($msg) . "\n\n";
            mbox_append($fn, $source);
            $app->log(qq{Saved message $id});
            $now++;
        };
        $app->log($@) if $@;
    }

    my $mbox = $now ? $fn : undef;
    my $flag = ($now == $total) ? 'good' : 'bad';

    my $note = q{};
    if($flag eq q{good}){
        $note = $loc->('Saved [quant,_1,message,messages]', $now);
    }else {
        $note = $loc->('Saved [_1] / [quant,_2,message,messages]', $now, $total);
    }

    return (
        %context,
        mbox => $mbox, 
        statusbar => $flag, 
        note => $note,
    );
}

sub _per_recipient_approve {
    my ($app, $to) = @_;
    my $msg_ids = $app->query->param('msgs');
    return unless $msg_ids;
    my $loc = $app->{context}{loc};

    my ($msg_id) = split ',', $msg_ids;
    my $q = PureMessage::Quarantine->new;

    my $bad_status = sub {
        my $status = shift;
        unless ($status) {
            my $msg = $q->message($msg_id);
            $status = get_message_status($msg);
        }
        return (
            message => { msg_id => $msg_id }, # for msg status baris
            statusbar => q{bad},
            note => $loc->('Approval failed for [_1]', $to),
            recip_status => $loc->('No action'),
            action_success => 0,
            status => { $msg_id => calc_message_status($status) },
        );
    };

    # We only pick up the first message_id based on the requirement.
    my $user = $app->{context}{user};
    if(defined $msg_id) {
        my $msg = $q->message($msg_id);
        my @recips = $msg->minfo_value('Envelope-To');

        # map the $to recipient to a group
        my $group_map = PureMessage::Group::CDBMap->Tie;
        my $to_group_name = $group_map->lookup($to);
        if (!$to_group_name and $user->super_user) {
            $to_group_name = 'supergroup';
        }
        elsif (!$to_group_name) {
            $app->log("Could not lookup group for user ($to) - cannot approve"
                      . " message $msg_id");
            return $bad_status->();
        }
        # check that that group is associated with $user
        my $have_perm = 0;
        my $to_group;
        for my $g ($user->groups) {
            next unless $to_group_name eq $g->group_name;
            $to_group = $g;
            last;
        }
        unless ($to_group) {
            $app->log("User is not associated with the '$to_group_name' group"
                      . " - cannot approve message $msg_id");
            return $bad_status->();
        }
        # check that $user/$group has approve permission
        unless ($user->can_do(group => $to_group_name, 
                              perm => 'quarantine.actions.approve')) {
            $app->log("User / group '$to_group_name' does not have approve "
                      . "permission for message $msg_id");
            return $bad_status->();
        }

        my $rc = approve([$msg], to => $to, logger => $q_logger,
                         quarantine => $q);
        my $status = get_message_status($msg);
        if ($rc) {
            $app->log(qq{Approved message $msg_id for $to});

            return (
                message => { msg_id => $msg_id }, # for msg status baris
                statusbar => q{good},
                note => $loc->('Approved message for [_1]', $to),
                recip_status => $loc->('Approved'),
                action_success => 1,
                status => {$msg_id => calc_message_status($status)},
            );
        }
        else {
            $app->log("Approval failed on message $msg_id for $to");
            return $bad_status->();
        }
    }
}

sub _handle_approve_delete {
    my ($app, $action) = @_;
    my $loc = $app->{context}{loc} || die "Can't find loc!";
    my %context;
    my $msg_list = $app->query->param('msgs');

    unless ($msg_list) {
        # XXX unlocalized
        return (statusbar => 'bad', note => 'No messages selected!');
    }
    my @msgs = split ',', $msg_list;
    my $total_msgs = @msgs;

    my $quar = PureMessage::Quarantine->new;
    my $took_action = 0; # number of messages changed
    for my $msg_id (@msgs) {
        my $msg = $quar->message($msg_id);
        unless (defined $msg) {
            $app->log("Couldn't create message $msg_id");
            $context{status}{$msg_id} = 'deleted';
            next;
        }

        # Check for permission to take the action
        my $have_perm = _have_permission($app, $action, $msg);
        unless ($have_perm) {
            $app->log("No permission to $action $msg_id");
            next;
        }

        # Don't take action on messages that have already been 
        # acted on
        my $status = get_message_status($msg);
        if ($action eq 'approve') {
            if ($status->{approve} == @{$status->{recips}}) {
                $app->log("Message $msg_id is already approved");
                $context{status}{$msg_id} = 'approved';
                next;
            }
            elsif ($status->{delete} == @{$status->{recips}}) {
                $app->log("Message $msg_id was already deleted by all users");
                $context{status}{$msg_id} = 'deleted';
                next;
            }
            elsif ($status->{normal} == 0) {
                $app->log("Message $msg_id was already approved or deleted by all users");
                $context{status}{$msg_id} = 'delete_approve';
                next;
            }
        }

        # If user can delete messages for all recipients, 
        # just call trash(), otherwise, must call trash($msg, to => $recip)
        # for each recipient the user can delete messages for
        my $group_map = PureMessage::Group::CDBMap->Tie;
        my $user = $app->{context}{user};
        my $is_superuser = $user->super_user;
        my @user_groups = map { $_->group_name } $user->groups;
        my $trasher = sub {
            my ($msgs, %opts) = @_;
            my $msg = $msgs->[0];
            my @recips = $msg->minfo_value('Mapped-Recipients');
            if ($msg->is_outbound) {
                # only one sender, so we know it's our group
                return trash([$msg], %opts, to => \@recips, tag_and_trash => 1);
            }
            # inbound: check recips, tag for delete or trash
            my %can_delete_for;
            for my $r (@recips) {
                # lookup recipient group, see if user manages that group
                my $recip_group = $group_map->lookup($r);
                if (!$is_superuser) {
                    next unless $recip_group;
                    next unless grep { $recip_group eq $_ } @user_groups;
                }
                next unless $user->can_do(group => $recip_group, 
                                          perm => 'quarantine.actions.delete');
                $can_delete_for{$r}++;
            }
            # user owns all recipients, so just trash it
            my @delete_for = keys %can_delete_for;
            if (@delete_for == @recips) {
                $app->log("All recipients managed by user's groups - "
                          . "deleting message $msg_id");
                return trash([$msg], %opts, to => \@recips, tag_and_trash => 1);
            }
            # mark message for delete for users of our groups
            my $trash_failed = 0;
            for my $recip (@delete_for) {
                my $rc = trash([$msg], %opts, to => $recip);
                if ($rc) {
                    $app->log("Marked message $msg_id for delete for "
                              . "recipient $recip");
                }
                else {
                    $trash_failed++;
                    $app->log("Failed to mark message $msg_id for delete "
                              . "for recipient $recip");
                }
            }
            return !$trash_failed;
        };

        my @recips = $msg->minfo_value('Mapped-Recipients');
        my $rc = ($action eq 'approve')
            ? approve([$msg], to => \@recips, logger => $q_logger, quarantine => $quar)
            : $trasher->([$msg], logger => $q_logger, quarantine => $quar);
        if ($rc) {
            $status->{$action}++;
            delete $status->{normal};
            delete $status->{approve} if $action eq 'delete';
            $took_action++;
            $app->log(ucfirst($action) . "d message $msg_id");
        }
        else {
            $app->log("Failed to $action message $msg_id");
        }

        $context{status}{$msg_id} = calc_message_status($status);
    }

    my $failed = $total_msgs - $took_action;
    my $actioned = ucfirst($action) . 'd';
    if ($failed) {
        $context{statusbar} = 'bad';
        $context{note} = $loc->("$actioned [_1] / [quant,_2,message,messages]",
                                $took_action, $total_msgs);
    }
    else {
        $context{statusbar} = 'good';
        $context{note} = $loc->("$actioned [quant,_1,message,messages]",
                                $took_action);
    }

    # needed by message popup status bar
    $context{message} = { msg_id => $msgs[0] };

    return %context;
}

sub _handle_delete_all {
    my ($app) = @_;
    my $loc = $app->{context}{loc};

    my $where = $app->query->param('where');
    # * means wide open search
    $where = '' if $where eq '*';

    my $user = $app->{context}->{user};
    my $groups;
    $groups = search_reasons_per_group($app);

    my %context;
    my $num_deleted = 0;
    my $total_msgs;
    eval {
        my $results = mset_query($where, $groups);
        $total_msgs = @{$results->{rows}};
        for my $row (@{$results->{rows}}) {
            my $msg_id = $row->message_id();
            my $msg = $results->{mstore}->message($msg_id);

            # Check for permission
            my $have_perm = _have_permission($app, 'delete', $msg);
            unless ($have_perm) {
                $app->log("No permission to delete $msg_id");
                next;
            }

            my @recips = $msg->minfo_value('Mapped-Recipients');
            my $rc = trash([$msg], logger => $q_logger, to => \@recips, tag_and_trash => 1);
            if ($rc) {
                $app->log("Deleted message $msg_id");
                $num_deleted++;
                $context{status}{$msg_id} = 'deleted';
            }
            else {
                $app->log("Failed to delete message $msg_id");
            }
        }
    };
    if ($@) {
        $app->log($@);
    }

    if ($total_msgs == $num_deleted) {
        $context{statusbar} = 'good';
        $context{note} = $loc->('Deleted [quant,_1,message,messages]',
                                $num_deleted);
    }
    else {
        $context{statusbar} = 'bad';
        $context{note} = $loc->('Deleted [_1] / [quant,_2,message,messages]',
                                $num_deleted, $total_msgs);
    }

    return %context;
}

sub _have_permission {
    my ($app, $action, $msg) = @_;
    my $user = $app->{context}{user};
    my $msg_groups = $msg->message_groups;
    my $have_perm = 0;
    for my $g ($user->groups) {
        # skip this users group if it is not associated with this message
        if (!$user->super_user) {
            next unless grep { $g->group_name eq $_ } @$msg_groups;
        }
        # check that this user/group can take action on the message
        next unless $user->can_do(group => $g->group_name,
                                  perm => "quarantine.actions.$action");
        $have_perm++;
    }
    return $have_perm ? 1 : 0;
}

1;
