package PureMessage::WebUI::Controller::Configuration;
use strict;
use warnings;
use ActiveState::Handy qw/cat/;
use PureMessage::Version qw($PREFIX);
use PureMessage::Group::PermissionItem;
use PureMessage::Group::PolicyOption;
use PureMessage::Resources::Manager;
use PureMessage::WebUI::Model::ConfigItem;

use base qw(Exporter);
our @EXPORT_OK = qw(handle_config_tab);

sub handle_config_tab {
    my ($app) = @_;
    my $context = $app->{context};
    $context->{tab} = q{config};
    my $user = $context->{user};
    # need to stringify $_->group_name to avoid lazy loading problem
    my @group_names = grep { $_ ne 'supergroup' }
                        map { '' . $_->group_name } 
                            $user->groups;

    update_current_group($app, \@group_names);

    # load data for sidebar
    $context->{config}{groups} = [
        sort
            grep { $user->can_do(group => $_, perm => 'configuration') }
                @group_names
    ];
    my $cur_group = $app->session->param('current_group');
    unless ($cur_group) {
        $cur_group = ${ $context->{config}{groups} }[0];
        $app->session->param('current_group', $cur_group);
    }
    $context->{config}{current_group} = $cur_group;
    my %items = (
        document => PureMessage::WebUI::Model::Document->new(app => $app),
        banner   => PureMessage::WebUI::Model::Banner->new(app => $app),
        list     => PureMessage::WebUI::Model::Multilist->new(app => $app),
    );
    for my $k (keys %items) {
        $context->{config}{$k} = $items{$k}->item_hash;
    }

    # Check for permissions
    my %perms = (policysettings    => 'can_read');
    for my $perm (keys %perms) {
        my $method = $perms{$perm};
        $context->{perm}{config}{$perm} = $user->$method(
                group => $cur_group,
                perm => "configuration.$perm"
            );
    }

    # Load the appropriate content for config page
    my $path_info = $app->query->path_info();
    if ($path_info =~ m#^/config/(\w+)/([^/]+)$#) {
        my $model = $items{$1} || die "No such item handler: $1";
        $items{$1}->process_item(id => $2);
        $context->{page} = $1;
    }
    elsif ($path_info =~ m#^/config/settings$#) {
        handle_settings($app, $cur_group);
        $context->{page} = q{settings};
    }
    else {
        # show default content
        $context->{page} = q{default};
        $context->{tmpl} = 'config/homepage.tmpl';
    }
}

sub handle_settings {
    my ($app, $cur_group) = @_;
    my $context = $app->{context};
    my $user = $context->{user};
    my $query = $app->query;

    # Load policy options from DB
    my @options = PureMessage::Group::PolicyOption::OptIn->retrieve_all;
    $context->{options} = [];
    my $can_write = 0;
    my $save_settings = $query->param('save_settings') || 0;
    my $changed = 0;
    for my $opt (@options) {
        my $name = $opt->name;
        my $perm = "configuration.policysettings.$name";
        $perm = $user->get_permissions(group => $cur_group, perm => $perm);
        next if $perm->value eq 'off';

        my $resource = _get_resource($name);
        unless ($resource) {
            $app->log("Cannot create resource '$name'");
            next;
        }
        my $matcher = $resource->native->matcher;

        if ($perm->value eq 'read/write') {
            $can_write++;
            if ($save_settings) {
                my $in_list = $matcher->($cur_group);
                my $new_value = $query->param("opt_$name") || '';
                my $list = $opt->list_name;
                if (defined $new_value) {
                    if ($opt->opt_in and $new_value) {
                        if (!$in_list) {
                            _add_to_list($app, $cur_group, $list, $resource);
                            $changed++;
                        }
                    }
                    elsif ($opt->opt_in) {
                        if ($in_list) {
                            _remove_from_list($app, $cur_group, $list, $resource);
                            $changed++;
                        }
                    }
                    elsif ($new_value) {
                        if ($in_list) {
                            _remove_from_list($app, $cur_group, $list, $resource);
                            $changed++;
                        }
                    }
                    else {
                        if (!$in_list) {
                            _add_to_list($app, $cur_group, $list, $resource) if !$in_list;
                            $changed++;
                        }
                    }
                }
            }
        }

        # determine checkbox status
        my $checked;
        # get a new matcher to avoid caching
        $matcher = $resource->native->matcher;
        if ($opt->opt_in) {
            $checked = $matcher->($cur_group);
        }
        else { # opt-out
            $checked = !$matcher->($cur_group);
        }

        $context->{perm}{config}{option}{$name} = $perm;
        push @{$context->{options}}, {
            option => $opt,
            checked => $checked,
        };
    }

    if ($changed) {
        $context->{statusbar} = 'good';
        $context->{note} = $context->{loc}->('Saved policy settings');
    }

    $context->{status_bar}{show_save} = $can_write;
    $context->{tmpl} = 'config/policy_settings.tmpl';
}

sub _add_to_list {
    my ($app, $group, $listname, $resource) = @_;
    $app->log("Adding group '$group' to $listname");
    $resource->cinsert(undef, $group);
    $resource->save;
}

sub _remove_from_list {
    my ($app, $group, $listname, $resource) = @_;
    $app->log("Removing group '$group' from $listname");
    $resource->cdelete(undef, $group);
    $resource->save;
}

sub _get_resource {
    my ($name) = @_;
    my $store = PureMessage::Resources::Store::DB->new;
    my $mgr = PureMessage::Resources::Manager->new(store => $store);
    return $mgr->find($name);
}

sub update_current_group {
    my ($app, $group_names) = @_;
    my $user = $app->{context}{user};

    # update the session with the new group, only if we admin that group
    if (my $new_group = $app->query->param('current_group')) {
        die "$new_group" if ref($new_group);
        if (grep { $new_group eq $_ } @$group_names) {
            if ($user->can_do(group => $new_group, perm => 'configuration')) {
                $app->session->param('current_group', $new_group);
            }
            else {
                $app->log("Can't set config group to $new_group - "
                          . "no permission");
            }
        }
        else {
            $app->log("Can't set config group to $new_group - not associated"
                      . " with user");
        }
    }

}
1;
