package PureMessage::WebUI::Controller::Report;
use strict;
use warnings;
use URI::Escape;
use Apache::Constants qw(SERVER_ERROR);
use PureMessage::WebUI::Utils qw(validate_start_end_dates $DEFAULT_DATE);
use PureMessage::WebUI::RunModeException;
use Sophos::Reports::ReportFactory;
use Sophos::Reports::Parameters qw(parameters_to_hash);
use Apache::Constants qw(FORBIDDEN);
use Data::Dumper;
use JSON::Syck;
use PureMessage::Reports::Custom ();

use base qw(Exporter);
our @EXPORT_OK = qw(handle_report);

# Handles report requests
sub handle_report {
    my ($app, $type) = @_;
    $app->{context}{tab}  = 'report';
    $app->{context}{page} = 'default';

    _parse_report_uri($app);
    my $custom_rpts = PureMessage::Reports::Custom::get_report_data();
    _load_report_permissions($app, $custom_rpts);

    if ($type eq 'asset') {
        eval { _process_report_asset($app) };
        if ($@) {
            my $msg;
            if (UNIVERSAL::isa($@, 'PureMessage::WebUI::AccessException')) {
                $msg = $@->error_message;
            }
            else {
                $msg = "Generating report image: $@";
            }
            $app->log($msg);
            $app->http_return_code(SERVER_ERROR);
            $app->header_props(-type => "text/plain");
            return $msg;
        }
        return $app->{context}{asset};
    } else {
        _load_custom_report_list($app, $custom_rpts);
        _process_report($app);
        return $app->tt_process(q{reports.tmpl}, $app->{context});
    }
}

sub _process_report_asset {
    my ($app) = @_;
    my $context = $app->{context};
    my $output = $context->{opts}{output};
    _check_perms($context, 'asset', $output);

    my $rep = _create_report_object($app, 'asset');
    my $report_id = $rep->report_id;
    if (defined $report_id) {
        $context->{report_obj} = $rep;
        $context->{report_id} = $report_id;
        $rep->set_groups($context->{reports}{groups})
            unless ($context->{user}->super_user()
                and $context->{opts}{only_group}
                and $context->{opts}{only_group} eq '__NO_GROUP__');

        if ($output eq 'csv') {
            my $filename = $rep->pretty_filename($output);
            $filename =~ s/:/_/g;
            $app->header_props(-type => 'text/comma-separated-values',
                              '-Content-Disposition' => "attachment; filename=$filename");
            $context->{asset} = $rep->csv;
        }
        elsif ($output eq 'txt') {
            $app->header_props(-type => 'text/plain');
            $context->{asset} = $rep->csv;
        }
        elsif ($output eq 'dat') {
            $app->header_props(-type => 'text/plain');
            $context->{asset} = Dumper $rep->data;
        }
        elsif ($output eq 'json') {
            $app->header_props(-type => 'text/plain');
            $context->{asset} = JSON::Syck::Dump($rep->data);
        }
        else {
            if (UNIVERSAL::can('GD::Image', $output)) {
                $app->header_props(-type => "image/$output");
                $context->{asset} = $rep->image($output);
            } else {
                die "Unknown output type: $output";
            }
        }
    } else {
        die "Unable to process report";
    }
    $app->log("Exporting report $report_id as $output");
}

sub _process_report {
    my ($app) = @_;
    my $context = $app->{context};
    my $report_id = $context->{opts}{report_id};
    if (defined $report_id) {
        $context->{report_id} = $report_id;
        $context->{report_template} = 'reports/' . $report_id . '.tmpl';
        _check_perms($context, 'table', 'html');
        # load the CGI vars and fix them up for passing to report object
        eval { _cleanup_report_options($app, 'html') };
        if ($@) {
            $context->{statusbar} = 'warning';
            $context->{note} = $@;

            # Create a default report object.
            # We need a default report obj so we can still present direction
            # and other ui details.

            # catch date range exception and display it on status bar
            if (UNIVERSAL::isa($@, 'PureMessage::WebUI::ValidationException')) {
                $context->{note} = $@->error_message;
                if ($@->{error_type} eq 'date_range') {
                    delete $context->{opts}{date_range_start};
                    delete $context->{opts}{date_range_end};
                }
            }
            eval {
                _cleanup_report_options($app , 'html');
                $context->{report_obj} = _create_report_object($app, 'html');
            };
            if ($@) {
                $context->{report_obj} = {};
                $context->{statusbar} = 'warning';
                $context->{note} = $@;
            }
        }
        $context->{report_obj} = _create_report_object($app, 'html');
    } else {
        # disable option bar, etc
        $context->{optionbar_disabled} = 1;
        # display default report landing page
        $context->{report_template} = 'reports/homepage.tmpl';
    }
}

sub _check_perms {
    my ($context, $type, $output) = @_;
    my $perms = $context->{perms};
    my $report_id = $context->{opts}{report_id};
    my $cst_rpt_key = $context->{opts}{cst_rpt_key} if exists $context->{opts}{cst_rpt_key};
    my $user = $context->{user};
    my $opts = $context->{opts};

    my $report_to_check;
    if (exists $context->{opts}{mkey}) {
        if (defined $cst_rpt_key) {
            $report_to_check = $cst_rpt_key;
        } else {
            # mkey but no cst_rpt_key means, no perm for this report
            die PureMessage::WebUI::AccessException->new(perm => "Policy Report");
        }
    } else {
        $report_to_check = $report_id;
    }

    die PureMessage::WebUI::AccessException->new(perm => "$report_to_check report")
        unless ($perms->{$report_to_check});

    $context->{reports}{groups} = _groups_that_can($context->{user}, $report_to_check);
    # Maintain backwards compatability... yes, this is confusing...
    if ($type eq 'asset' && $output eq 'csv') {
        my $report_key = defined $cst_rpt_key ? $cst_rpt_key : $report_id;
        my $export_key = "$report_key.export";
        die PureMessage::WebUI::AccessException->new(perm => "$report_key report")
            unless ($perms->{$export_key});
        if (!$user->super_user and !($opts->{only_group} and $opts->{only_group} eq '__NO_GROUP__')) {
            $context->{reports}{groups}= _groups_that_can($user, "$report_to_check.export");
        }
    }

    die PureMessage::WebUI::AccessException->new(perm => "$report_to_check report")
        if (!@{$context->{reports}{groups}} && !$user->super_user());
}

sub _load_custom_report_list {
    my ($app, $custom_rpts) = @_;
    my $user  = $app->{context}{user};
    my $perms = $app->{context}{perms};

    # only check individual report permission when user has perm to "custommarkhits"
    # use the context->{perms} previously created by _load_report_permission()
    return unless $perms->{'CustomMarkHits'};

    foreach my $k (sort { $a <=> $b } keys %{$custom_rpts}) {
        my $rep_name = "CustomMarkHits." . $custom_rpts->{$k}->{key};
        push @{$app->{context}{custom_report_list}}, {
            title => $custom_rpts->{$k}->{title},
            id    => $k,
        } if $perms->{$rep_name};
    }
}

sub _load_report_permissions {
    my ($app, $custom_rpts) = @_;
    my $report_id   = $app->{context}{opts}{report_id};
    my $mkey        = $app->{context}{opts}{mkey};

    my $user = $app->{context}{user};
    $app->{context}{perms} = ();

    # XXX These report names should not be hardcoded.
    my @reports = qw(
        MessageCategories MessagesBlocked TopVirusTypes TopSpamRelays 
        TopVirusRelays TopOtherRelays TopRelays QueueStatus TopSpamRecipients 
        TopSpamSenders TopVirusSenders CategoriesByGroup MtaRejected PolicyMarkHits
        CustomMarkHits);

    $app->{context}{perms}{'reportexport'}      = 0;
    for my $rep (@reports) {
        # Add 'report' permission to the list
        $app->{context}{perms}{$rep} = $user->any_group_can(lc "reports.$rep");
        # Add 'export' permission to the list
        if (defined $report_id && $rep eq $report_id && $rep ne 'CustomMarkHits') {
            my $rep_export = "$rep.export";
            $app->{context}{perms}{$rep_export} = $user->any_group_can(lc "reports.$rep_export");
            $app->{context}{perms}{'reportexport'} =  $app->{context}{perms}{$rep_export};
        }
    }

    if ($app->{context}{perms}{'CustomMarkHits'}) {
        foreach my $k (keys %{$custom_rpts}) {
            my $rep_name = "CustomMarkHits." . $custom_rpts->{$k}->{key};
            my $rep_perm_name = "reports.$rep_name";
            my $any_group_can = 0;
            eval { $any_group_can = $user->any_group_can(lc $rep_perm_name) };
            if ($@ || !$any_group_can) {
                $app->{context}{perms}{$rep_name} = 0;
            } else {
                $app->{context}{perms}{$rep_name} = 1;
                # check current custom report's export perm
                if (defined $mkey && $k == $mkey) {
                    $app->{context}{opts}{cst_rpt_key} = $rep_name;
                    my $export_name  = "$rep_perm_name.export";
                    my $any_group_can_export = 0;
                    eval { $any_group_can_export = $user->any_group_can(lc $export_name) };
                    if ($@ || !$any_group_can_export) {
                        $app->{context}{perms}{"$rep_name.export"} = 0;
                        $app->{context}{perms}{'reportexport'}  = 0;
                    } else {
                        $app->{context}{perms}{"$rep_name.export"} = 1;
                        $app->{context}{perms}{'reportexport'}  = 1;
                    }
                }
            }
        }
    }
}

sub _parse_report_uri {
    my $app = shift;
    my $query = $app->query;
    my $path_info;
    $app->{context}{opts} = { %{ $query->Vars } };
    $app->{context}{opts}{output} = 'html';
    # determine if asset
    if ($query->path_info =~ m#^/reports/?#) {
        ($path_info = $query->path_info) =~ s#^/reports/?##;
        $app->{context}{opts}{'is_asset'} = 0;

    } elsif ($query->path_info =~ m#^/report-img/#) {
        ($path_info = $query->path_info) =~ s#^/report-img/##;
        $app->{context}{opts}{'is_asset'} = 1;

        if ($path_info =~ m#\w+/(.*)#) {
            $app->{context}{opts}{report_uri} = $1;
            my ($parameters, $output) = $app->{context}{opts}{report_uri} =~ m#^([^.?]+)(?:\.([^/]*))?#;
            $app->{context}{opts}{output} = $output if defined $output;
        }
    }

    if ($path_info =~ m#(\w+)#) {
        $app->{context}{opts}{report_id} = $1;
    }
}

sub _create_report_object {
    my ($app, $type) = @_;
    my $rep;
    my $opts = $app->{context}{opts};
    my $user = $app->{context}{user};
    my $report_group = delete $opts->{report_group} || '';
    if ($user->super_user and $report_group eq '_all_mail') {
        $opts->{only_group} = '__NO_GROUP__';
    }
    $opts->{user} = $user;
    # if super_user and only_group eq '__NO_GROUP__ dont set groups
    $opts->{groups} = $app->{context}{reports}{groups}
        unless ($user->super_user and $opts->{only_group} and $opts->{only_group} eq '__NO_GROUP__');
    $opts->{logger} = sub { $app->log(shift) };

    if ($type eq 'asset') {
        $rep = Sophos::Reports::ReportFactory->create(%{$opts});
    } else {
        my $report_id = $opts->{report_id};
        my $class = "Sophos::Reports::Report::$report_id";
        eval "require $class;";
        if ($@) {
            $app->log($@);
            die if $@;
        }
        $rep = $class->new(%{$opts});
    }
    return $rep;
}

sub _default_report_opts {
    my $report_id = shift;

    my $defaults_map = {
        'policy-blocked' => {
            'history' => q{12h},
            'format' => q{bars_stacked},
        },
        topspamrelays => {
            history => q{12h},
            format  => q{tbl},
        },
    };

    return exists($defaults_map->{$report_id})
        ? %{$defaults_map->{$report_id}} : ();
}

sub _calc_report_opts {
    my $report_id = shift;

    my %opts = (
        table_view => 1,
        bar_view   => 1,
        line_view  => 1,
        direction_view => 0,
        view => 'bars_stacked',
        date_range_start => $DEFAULT_DATE,
        date_range_end => $DEFAULT_DATE,
    );

    if ($report_id eq 'policy-blocked') {
        $opts{line_view} = 0;
        $opts{direction_view} = 0;
    }
    return \%opts;
}

sub _cleanup_report_options {
    my ($app, $output) = @_;
    my $context = $app->{context};
    my $opts    = $app->{context}{opts};

    if (exists $opts->{format}) {
        $opts->{type} = $opts->{format};
        delete $opts->{format};
    }

    my $user = $context->{user};
    if (!$opts->{report_group}) {
        $opts->{report_group} = '_all_mail' if $user->super_user;
    }
    elsif ($opts->{report_group} eq '_all_groups') {
        $context->{reports}{selected} = 'all_groups';
    }
    elsif ($user->super_user and $opts->{report_group} eq '_all_mail') {
        $context->{reports}{selected} = 'all_mail';
    }

    # clear unset options
    for (qw(direction history)) {
        delete $opts->{$_} if $opts->{$_} and $opts->{$_} eq '-';
    }

    if ($opts->{history}) {
        $opts->{span} = $opts->{history};
        # date will be set by Sophos::Reports::PMX
        if ($opts->{span} =~ /sfn=(\d)/) {
            # if sfn (start from now) is defined and its value is none
            # zere then the end of search period is now.
            $opts->{sfn} = $1;
            $opts->{span} =~ s/\s+sfn=\d//;
        }
    }
    elsif ($opts->{date_range_start} || $opts->{date_range_end}) {
        _cleanup_datetime_options($app, $context, $opts);
    }

    if (my $group = $opts->{report_group}) {
        if (grep { $_ eq $group } @{$context->{reports}{groups}}) {
            $opts->{only_group} = $group;
        }
    }

    $context->{default_date} = $DEFAULT_DATE;
}

sub _cleanup_datetime_options {
    my ($app, $context, $opts) = @_;

    my $loc = $context->{loc};
    my ($date, $span, $end_date) = validate_start_end_dates($loc,
                                        start_date_mandatory => 1,
                                        start_date => $opts->{date_range_start},
                                        end_date => $opts->{date_range_end});
    $opts->{date} = $date;
    $opts->{span} = "$span minutes" if defined $span;
    $opts->{date_range_end} = $end_date;
}

sub _groups_that_can {
    my ($user, $report_id) = @_;
    my @groups = ();
    foreach my $g ($user->groups) {
        my $group = $g->group_name;
        next if $group eq 'supergroup';
        if ($user->can_do(group => $group, perm => lc("reports.$report_id"))) {
            push(@groups, "$group");
        }
    }
    @groups = sort @groups;

    return \@groups;
}

###
# END
1;
