package PureMessage::WebUI::Controller::Search;
use strict;
use warnings;
use URI::Escape;
use HTTP::Date;
use POSIX qw(strftime mktime);
use ActiveState::I18N::UTF8 qw(utf8_header);
use PureMessage::Quarantine;
use PureMessage::UTF8::Config qw(%utf8_config);
use PureMessage::QDigest::MessageId qw(extract_ids decode_id);
use PureMessage::WebUI::Utils qw(htmlify get_message_status $DEFAULT_DATE
                                 message_status_from_id mset_query
                                 validate_start_end_dates parse_date);
use PureMessage::Group::PermissionItem;
use PureMessage::MessageStore::pmdb::MessageSet; # For $REASON_ABLE flag

use base 'Exporter';
our @EXPORT_OK = qw(handle_search load_search_reasons search_reasons_per_group);

use constant DEFAULT_SORT_ORDER => scalar(q{desc});
use constant DEFAULT_SORT_BY => scalar(q{date});
use constant DEFAULT_START_DATE => q{2000-01-01 00:00};

# Set the REASON_ABLE flag, indicating that we want to use the message.reason
# field instead of joining the message table with m_reason table.
# When the m_reason table is dropped, this flag can be removed.
PureMessage::MessageStore::pmdb::MessageSet::set_reason_able(1);

my %_SORT_MAP = (
    date => 'c_qdate_time',
    reason => 'm_reason',
    recipient => 'h_to',
    sender => 'm_envelope_from',
    subject => 'c_subject_utf8',
);

sub handle_search {
    my ($app) = @_;
    my $context = $app->{context};

    $context->{tab} = q{search};
    $context->{page} = q{default};

    my %search_details;
    
    $search_details{$_} = $context->{search}->{$_} foreach keys(%{$context->{search}});
    my $results;

    for my $key (qw(sender recipient subject reason qdateFrom qdateTo relay queue_id)){
        $search_details{params} ||= {};
        $search_details{params}{$key} = $app->query->param($key);
    }
    
    eval {
        my $loc = $context->{loc};
        $results = perform_search($app, $context);
        $search_details{results} = $results->{results};
        $search_details{page} = $results->{page};
        if ($search_details{page}{number}) {
            $search_details{page}{display} = 
                $loc->('Page [_1] of [_2]', $search_details{page}{number},
                                            $search_details{page}{total});
        }

        # we'll be posting the where string back for 'delete all'
        # so make sure it's properly escaped
        $search_details{where} = $results->{where} || q{*}; # * means wide open
        $search_details{where} = uri_escape($search_details{where});
        $search_details{reason} = $results->{reason} || '';
        # put search details into template
        my $num = $results->{estimated};
        $context->{statusbar} = $num > 0 ? 'good' : 'warning';
        my $many = $num == 1 ? q{} : q{s};        
        if ($results->{approx}) {
            $context->{note} = $loc->('Your search returned approximately [quant,_1,result,results].', $num);
        }
        else {
            $context->{note} = $loc->('Your search returned [quant,_1,result,results].', $num);
        }
    };
    if ($@) {
        my $where = $search_details{where} || '';
        $context->{statusbar} = q{warning};
        $context->{note} = $@;
        $search_details{results} = [];
        
         # catch date range exception and display it on status bar
        if (UNIVERSAL::isa($@, 'PureMessage::WebUI::ValidationException')) {
            $context->{note} = $@->error_message;
            if ($@->{error_type} eq 'date_range') {
                delete $search_details{params}{dateFrom};
                delete $search_details{params}{dateTo};
            }
        }
    }
    $context->{search} = \%search_details;
    # Save the date fields between submits. If we don't do this, the user 
    # cannot advance past the first page of a multi-page query since the date 
    # fields will be reset to their default value on each page reload.
    $context->{date_from} = $search_details{params}{dateFrom} 
        if defined($search_details{params}{dateFrom});
    $context->{date_to} = $search_details{params}{dateTo}
        if defined($search_details{params}{dateTo});
}

sub perform_search {
    my ($app, $context) = @_;

    # perform search
    my $where = _create_where($app);

    my $groups;
    my $user = $context->{user};

    $groups = search_reasons_per_group($app);
    $app->log("qsearch: '$where'");

    my $results;
    my $page_size   = $app->query->param(q{page_size}) || 0;
    $page_size = 0 unless $page_size =~ /^\d+$/;
    my $page_number = $app->query->param(q{page_number}) || 1;

    my $sort_by = $app->query->param(q{sort_by}) || DEFAULT_SORT_BY;
    $sort_by = DEFAULT_SORT_BY unless exists $_SORT_MAP{$sort_by};

    my $sort_order = $app->query->param(q{sort_order}) || DEFAULT_SORT_ORDER;
    $sort_order = DEFAULT_SORT_ORDER
        unless ($sort_order eq q{asc} or $sort_order eq q{desc});

    my $order_by = $_SORT_MAP{$sort_by} . ' ' . $sort_order;

    if(defined $app->query->param(q{next_page})){
        $page_number++;
    }

    if(defined $app->query->param(q{prev_page})){
        $page_number-- unless $page_number == 1;
    }

    if(defined $app->query->param(q{first_page})){
        $page_number = 1;
    }

    if(defined $app->query->param(q{last_page})){
        $page_number = 999_999_999_999_999;
    }

    eval {
        $results = mset_query($where, $groups, $page_size, $page_number, $order_by);
        if($results->{page_number} != $page_number){
            $page_number = $results->{page_number};
        }
    };
    if ($@) {
        warn $@;
        $app->log($@);
        # nice user visible error message
        die $context->{loc}->('An error occurred while performing your search');
    }

    my @results;
    my $i = 0;
    for my $row (@{$results->{rows}}) {
        my $info = _clean_mset_row($row);
        $info->{num} = $i++;
        $info->{status} = message_status_from_id($results->{mstore},
                                                 $info->{message_id});
        push @results, $info;
    }

    my $total_pages = $results->{total_pages};
    my $can_page_prev = $page_number > 1 ? 1 : 0;
    my $can_page_next = $page_number < $total_pages ? 1 : 0;
    my $can_page_first = $page_number != 1 ? 1 : 0;
    my $can_page_last = $page_number != $total_pages ? 1 : 0;

    my $reason = $app->query->param('reason');
    return { page => {
                     total      => $total_pages,
                     size       => $page_size,
                     can_prev   => $can_page_prev,
                     can_next   => $can_page_next,
                     can_first  => $can_page_first,
                     can_last   => $can_page_last,
                     number     => $page_number,
                     sort_by    => $sort_by,
                     sort_order => $sort_order,
                },
             results => \@results,
             where => $where,
             reason => $reason,
	     estimated => $results->{estimated},
    	     approx => $results->{approx},
           };
}

sub _create_where {
    my $app = shift;

    my $start_date = $app->query->param('qdateFrom')  || DEFAULT_START_DATE;
    $start_date = DEFAULT_START_DATE if $start_date eq $DEFAULT_DATE;
    my $end_date = $app->query->param('qdateTo');
    validate_start_end_dates($app->{context}{loc},
                             start_date => $start_date, 
                             end_date => $end_date);

    my @clauses;
    push @clauses,
            _make_query($app, 'c_from', 'sender'),
            _make_query($app, 'c_to', 'recipient'),
            _make_default_query($app, 'c_subject_utf8', 'subject'),
            _make_default_query($app, 'm_relay', 'relay'),
            # XXX - This works... however it's STUPIDLY crufty
            # &validate_start_end_dates needs to deal with a non-existing
            # start_date in certain scenarios.
            ($start_date eq DEFAULT_START_DATE 
                ? () 
                : _make_date_query($app,'From', $start_date)
            ),
            _make_date_query($app, 'To', $end_date),
            _query_queue_id($app);

    my $where = join( ' AND ',
                     grep {defined $_ && ($_ !~ /^\s*\(?\s*\)?\s*$/)}
                     @clauses );
    return $where;
}

sub _query_queue_id {
    my $app = shift;
    my $id = $app->query->param('queue_id');
    return unless $id;

    $id = fixup_pattern($id);

    # qdigest-encoded store id
    my @ids = extract_ids( $id );
    if (@ids) {
        $id = decode_id($ids[0], PureMessage::Quarantine->new);
        return "m_global_id == '$id'";
    }
    # numerical store id
    elsif ($id =~ /^\d+$/) {
        return "id == '$id'";
    }
    # h_message_id
    elsif ($id =~ /^<.*\@.*>$/) {
        return "h_message_id == '$id'";
    }
    # m_global_id (xxxx-xx)
    elsif ($id =~ /^\d+-\d+$/) {
        return "m_global_id == '$id'";
    }
    # at least one non-whitespace please (mta queue id)
    elsif ($id =~ /\S/) {
        return "m_queue_id == '$id'";
    }
    # empty field
}

sub fixup_pattern {
    my $pattern = shift;

    # map * wildcards to %s
    $pattern =~ s#\*#%#g;

    # remove forbidden characters
    $pattern =~ s#'#''#g;

    # remove leading and trailing whitespace
    $pattern =~ s#^\s+##;
    $pattern =~ s#\s+$##;

    return $pattern;
}

sub _make_query {
    my ($app, $msg_field, $cgi_field) = @_;
    my $pattern = lc( $app->query->param($cgi_field) );
    return unless $pattern;

    $pattern = fixup_pattern($pattern);

    if ($pattern =~ /\%/) {
        return "any $msg_field like '$pattern'";
    }
    # do exact match
    return "any $msg_field = '$pattern'";
}

sub _make_default_query {
    my ($app, $msg_field, $cgi_field, $exact) = @_;
    my $pattern = $app->query->param($cgi_field);
    return unless $pattern;

    $pattern = fixup_pattern($pattern);

    return "$msg_field == '$pattern'" if $exact;
    return "any $msg_field like '\%$pattern\%'";
}

sub _make_date_query {
    my ($app, $type, $date) = @_;
    return unless $date;
    return if $date eq $DEFAULT_DATE;
    my ($year, $month, $day, $hour, $min, $sec);

    $date = fixup_pattern($date);

    my $loc = $app->{context}->{loc};
    my $is_end_date = $type ne 'From';
    ($year, $month, $day, $hour, $min, $sec) = 
            parse_date($date, $is_end_date)
        or die $loc->(qq{Invalid date format: [_1]\n}, $date);

    # Convert to 0-based counting.
    --$month;
    $year -= 1900;

    my $date_t = mktime($sec, $min, $hour, $day, $month, $year)
        or die $loc->(qq{Invalid date: [_1]\n}, $date);
    
    # Do a little DST-based guessing
    # This may be incorrect in some locales...
    my @rebuilt_date = localtime($date_t);
    $date_t -= 3600 if ($rebuilt_date[8]);
    
    my $comparator = $is_end_date ? '<' : '>';
    return qq{c_qdate_time $comparator $date_t};
}

sub _clean_mset_row {
    my $row = shift;

    # Grab the fields out of the result set.
    my ($reason, $from, $to, $date, $subj, $prob) = $row->fields();
    my $status;

    # CLEAN: Reason
    if (lc($reason) eq 'spam' and defined $prob) {
        $reason .= sprintf( ' %.1f%%', ($prob*100) );
    }
    $reason ||= '(unknown)';

    # CLEAN: From
    $from ||= '(none)';
    utf8_header( $from, \%utf8_config );

    # CLEAN: To
    $to ||= '(none)';
    utf8_header( $to, \%utf8_config );

    # CLEAN: Date
    $date = str2time($date);
    $date = time unless $date;

    # CLEAN: Subject
    $subj ||= '(none)';

    # Put all of the clean data into our info hash.
    my $clean = {
        date   => strftime('%a, %d %b %H:%M', localtime($date)),
        sender => $from,
        recipient => $to,
        subject => $subj,
        attachment => 0, # TODO
        reason => $reason,
        message_id  => scalar $row->message_id(),
    };

    # QUERY: status
#    $clean->{'status'} = $self->_get_message_status( $clean );


    # Make sure that HTML Entities are escaped properly in all of the
    # info items.
    foreach my $key (keys %{$clean}) {
        $clean->{$key . "_raw"} = $clean->{$key};
        $clean->{$key} = htmlify( $clean->{$key} );
    }

    return $clean;
}

sub load_search_reasons {
    my ($app) = @_;
    my $user = $app->{context}{user};
    my $base_perm = 'quarantine.reason.';
    my $reason_re = qr/^\Q$base_perm\E/;
    my $reason_perms = PureMessage::Group::PermissionItem::permission_list($reason_re);
    my @reasons;
    for my $r (@$reason_perms) {
        # Localize?
        (my $reason_name = $r) =~ s#\Q$base_perm\E##;
        push @reasons, ucfirst($reason_name) if $user->any_group_can($r);
    }
    return \@reasons;
}

sub search_reasons_per_group {
    my ($app) = @_;
    my $user = $app->{context}{user};
    my $base_perm = 'quarantine.reason.';

    my $reasons = [];
    my $reason = lc $app->query->param('reason');
    unless($reason) {
        my $reason_re = qr/^\Q$base_perm\E/;
        $reasons = PureMessage::Group::PermissionItem::permission_list($reason_re);
    }
    else {
        $reasons = ["$base_perm$reason"];
    }

    my %reasons;
    my @groups;
    if ($user->super_user) {
       push(@groups, 'supergroup');
    }
    else {
       @groups = map { $_->group_name } $user->groups;
    }
    for my $g (@groups) {
        for my $r (@$reasons) {
            # Localize?
            (my $reason_name = $r) =~ s#\Q$base_perm\E##;
            $reasons{$g}{$reason_name}++ if $user->can_do(group => $g, perm => $r);
        }
    }

    return \%reasons;
}

1;
