package PureMessage::WebUI::Model::ConfigItem;
use strict;
use warnings;
use PureMessage::WebUI::RunModeException;
use Carp;

sub new {
    my ($class, %opts) = @_;
    die "Must supply a webapp object!" unless $opts{app};
    my $self = { %opts };
    bless $self, $class;
    return $self;
}

sub _perm       { die "subclasses must implement" }
sub _template   { die "subclasses must implement" }
sub _item_name  { die "subclasses must implement" }

sub item_hash {
    my ($self) = @_;
    my $app = $self->{app};
    my $perm = 'configuration.' . $self->_perm . '.';
    my $list_perms = PureMessage::Group::PermissionItem::permission_list(qr/\Q$perm\E/);
    my @items;
    my $user = $app->{context}{user} || die "Can't find user!";
    my $cur_group = $app->{context}{config}{current_group} 
                    || die "Can't find current_group!";
    for my $l (@$list_perms) {
        (my $shortname = $l) =~ s#\Q$perm\E##;

        my $can_read;
        eval { $can_read = $user->can_read(group => $cur_group, perm => $l) };
        next unless $can_read;

        my $item;
        eval { $item = $self->item($shortname) };
        if ($@ or !$item) {
            $app->log("Unknown config item '$shortname': $@");
            next;
        }

        my $desc = undef;
        eval { $desc = $item->description; };
        
        my $hname = undef;
        eval { $hname = $item->hname; };
        if (!defined $hname || $hname eq $item->name) {
            $hname = $desc;
        }
 
        push @items, { 
            id    => $shortname, 
            name  => $item->name, 
            desc  => $desc, 
            hname => $hname,
        };
    }
    @items = sort { $a->{name} cmp $b->{name} } @items;
    return \@items;
}

sub process_item {
    my ($self, %opts) = @_;
    my $id = $opts{id} || croak "Item id is mandatory!";
    my $app = $self->{app};
    my $context = $app->{context} || die "No context!";
    my $cur_group = $context->{config}{current_group} || die "No current group!";

    # check that user has permission to read/write this list
    my $user = $context->{user};
    my $perm = "configuration." . $self->_perm . ".$id";
    unless ($user->can_read(group => $cur_group, perm => $perm)) {
        die PureMessage::WebUI::AccessException->new(perm => $perm, 
                                                     prohibited => 1);
    }

    my $item;
    eval { $item = $self->item($id) };
    if ($@) {
        my $msg = "Invalid item ($id)";
        $app->log("$msg: $@");
        die $msg;
    }
    my $title = $item->name;
    eval { $title = $item->description };
    eval { $title = $item->hname };

    # Save the updated list contents, if necessary
    my $can_write = $user->can_write(group => $cur_group, perm => $perm);
    my $new_content = $app->query->param('edit_box');
    if (defined $new_content) {
        if ($can_write) {
            $app->log("Saving new content for item $id, group $cur_group");

            $new_content =~ s/[\x00-\x09\x0B-\x1F]//sg;
            $self->replace_group_content(group => $cur_group, 
                                         item  => $item,
                                         content => $new_content);
            $context->{note} = $context->{loc}->('[_1] saved', $title);
            $context->{statusbar} = 'good';
        }
        else {
            $context->{note} = 
                $context->{loc}->('[_1] not saved - no permission', $title);
            $context->{statusbar} = 'bad';
        }
    }

    $context->{item} = { name => $title,
                         id => $id,
                         contents => $self->group_content($item, $cur_group),
                       };
    $context->{status_bar}{show_save} = $can_write;
    $context->{tmpl} = $self->_template;
}


package PureMessage::WebUI::Model::Multifile;
use strict;
use warnings;
use base 'PureMessage::WebUI::Model::ConfigItem';
use PureMessage::Multifile::DB;

sub item {
    my ($self, $id) = @_;
    return PureMessage::Multifile::DB->load($id);
}

sub replace_group_content {
    my ($self, %opts) = @_;
    my $mf      = $opts{item} || die 'item is mandatory';
    my $group   = $opts{group} || die 'group is mandatory';
    my $content = $opts{content};
    die "No content defined!" unless defined $content;
    $mf->set_content($group, $content);
};

sub group_content {
    my ($self, $mf, $group) = @_;
    return $mf->content($group);
}

package PureMessage::WebUI::Model::Document;
use strict;
use warnings;
use base 'PureMessage::WebUI::Model::Multifile';
use PureMessage::Multifile::DB;

sub _perm { 'documents' }
sub _template { 'config/document.tmpl' }

package PureMessage::WebUI::Model::Banner;
use strict;
use warnings;
use base 'PureMessage::WebUI::Model::Multifile';
use PureMessage::Multifile::DB;

sub _perm { 'banners' }
sub _template { 'config/banner.tmpl' }

package PureMessage::WebUI::Model::Multilist;
use strict;
use warnings;
use base 'PureMessage::WebUI::Model::ConfigItem';
use PureMessage::MultiList;
use PureMessage::Resources::Manager;

sub _perm { 'lists' }
sub _template { 'config/list.tmpl' }

sub _resource_manager {
    my $self = shift;
    unless ($self->{_rmgr}) {
        $self->{_rmgr} = PureMessage::Resources::Manager->new(writable => 1);
    }
    return $self->{_rmgr};
}

sub _get_resource {
    my ($self, $rname) = @_;
    my $rmgr = $self->_resource_manager();
    my $resource;   
    eval { $resource = $rmgr->find($rname) };
    die "can't find resource $rname: $@" if $@;
    unless ($resource) {
        eval { $resource = $rmgr->add($rname) };
        die "can't add resource $rname: $@" if $@;
    }
    return $resource;
}

sub item {
    my ($self, $shortname, %args) = @_;
    return $self->_get_resource($shortname);
}

sub replace_group_content {
    my ($self, %opts) = @_;
    my $group   = $opts{group} || die 'group is mandatory';
    my $list    = $opts{item} || die 'item is mandatory';
    my $content = $opts{content};
    die "No content defined!" unless defined $content;

    # save to the database    
    my $sublist = $self->_load_sublist($list, $group);
    my %unique_items;
    $unique_items{$_}++ for split("\n", $content);
    $sublist->data()->commit_clear();
    $list->cinsert($group, $_) for (keys %unique_items);
    $list->save();    
}

sub _load_sublist {
    my ($self, $list, $group) = @_;
    unless ($self->{_sublist}) {
        my $native_list = $list->native();
        $self->{_sublist} = $native_list->find_list(key => $group, 
                                                    create_mode => 1);
    }
    return $self->{_sublist};
}

sub group_content {
    my ($self, $list, $group) = @_;
    my $sublist = $self->_load_sublist($list, $group);
    my $data = $sublist->data();
    my @contents;
    $data->enumerate_items(sub { 
        push @contents, shift;
    });
    return join("\n", @contents);
}

1;

