package PureMessage::WebUI::Utils;
use strict;
use warnings;
use Time::Local ();
use POSIX qw(strftime mktime);
use ActiveState::DateTime qw(check_date);
use ActiveState::UTF8 qw(is_valid_utf8);
use ActiveState::I18N::UTF8 qw(utf8_length);
use base 'Exporter';
our @EXPORT_OK = qw(htmlify get_message_status mset_query
                    calc_message_status message_status_from_id
                    validate_start_end_dates parse_date $DEFAULT_DATE
                    utf8_truncate_filter);
use PureMessage::Quarantine;
use Carp qw(croak);

our $DEFAULT_DATE = 'YYYY-MM-DD [hh:mm]';

# Basic HTML-ification cleanups; transforms several commonly used characters
# into HTML-friendly versions (e.g. "&" => "&amp;").
#
# NOTE: It is recommended that modules use -THIS- method instead of
# 'HTML::Entities'; 'encode_entities()' isn't UTF-8 aware and it mangles UTF-8
# text accidentally.
sub htmlify {
    my $str = $_[0];
    if (defined $str) {
        $str =~ s/&/&amp;/g;
        $str =~ s/</&lt;/g;
        $str =~ s/>/&gt;/g;
        $str =~ s/\"/&quot;/g;
        $str =~ s/\'/&#39;/g;
        $str =~ s/\n/<br>\n/g;
    }
    return $str;
}

sub message_status_from_id {
    my ($mstore, $msg_id) = @_;
    croak("Must supply a message id!") unless $msg_id;
    my $msg = $mstore->message($msg_id);
    die "Couldn't create msg object ($msg_id)" unless defined $msg;
    my $status = get_message_status($msg);
    return calc_message_status($status);
}

sub get_message_status {
    my ($message) = @_;
    my @recips  = $message->minfo_value( 'Envelope-To' );
    my %actions = $message->user_actions( \@recips );
    my %default_status = (
        recips => \@recips,
        normal => 0,
        delete => 0,
        approve => 0,
    );
    my %status = %default_status;
    foreach my $action (values %actions) {
        if (!defined $action) {
            $status{normal}++;
        }
        elsif ($action =~ /approve/) {
            $status{approve}++;
        }
        elsif ($action =~ /delete/) {
            $status{delete}++;
        }
        else {
            warn "Unknown action: $action";
        }
    }

    # check for locally trashed message
    if(my $local_msg = $message->get_local_message){
        if ($local_msg->folder eq 'trash') {
            return {
                %default_status,
                'delete' => scalar(@recips),
            };
        }
    }else{
        # we need to check the quarantine 'release' queue for this message

        #
        # XXX - ADD UNIT TESTS FOR THIS!
        #
        my $release_queue = $message->message_store()->get_release_queue();
        my $location_name = ${ $message->locations() }[0];
        my $message_set = $release_queue->get_requests($location_name, q{delete});

        if (grep { $_->{global_id} eq $message->id } @$message_set) {
            return {
                %default_status,
                'delete' => scalar(@recips),
            };
        }
    }

    return \%status;
}

sub mset_query {
    my ($where, $groups, $page_size, $page_number, $order_by) = @_;
    my $mstore = PureMessage::Quarantine->new;

    my %opts = (
        fields => [qw(m_reason m_envelope_from h_to m_date
                      c_subject_utf8 m_spam_probability)],
        reverse => 1,
        distinct => 1,
        where => $where,
        ($groups && scalar(%$groups)) ? (groups => $groups) : (),
    );

    if ($order_by) {
        $opts{order_by} = (ref($order_by) eq 'ARRAY') ? $order_by : [$order_by];
    }

    my $mset;
    my @rows;
    my $total_pages = 0;
    my $estimated_results = 0;
    my $approx = 0;

    if ($page_size){
        $page_number ||= 1;
        # If we're paging we need to use 'fast_fetch'
        $mset = $mstore->message_set(%opts, execute => 0);

        # Get total number of pages for this page size
        ($estimated_results, $approx) = $mset->estimate_rows();
        $total_pages = ($estimated_results / $page_size);
        # round up!
        if($total_pages > int($total_pages) || $total_pages == 0){
            $total_pages = int $total_pages;
            $total_pages++;
        }

        # reset page_number to the last page if it's greater than total
        $page_number = $total_pages if($page_number > $total_pages);

        my $offset = (($page_number - 1) * $page_size);
        @rows = $mset->fast_fetch($offset, $page_size);

    }else {
        $mset = $mstore->message_set(%opts);
        my $i = 0;
        while (my $row = $mset->fetch($i++)) {
            $estimated_results++;
            push @rows, $row;
        }
    }

    return { rows => \@rows,
             mstore  => $mstore,
             total_pages => $total_pages,
             page_number => $page_number,
             estimated => $estimated_results,
             approx => $approx,
           };
}

sub calc_message_status {
    my $s = shift;
    # convert message status into icon name
    if ($s->{approve} and $s->{delete}) {
        return $s->{normal} ? 'multistatus' : 'delete_approve';
    }
    elsif ($s->{approve}) {
        return $s->{normal} ? 'part_approved' : 'approved';
    }
    elsif ($s->{delete}) {
        return $s->{normal} ? 'part_deleted' : 'deleted';
    }
    return "empty";
}

sub validate_start_end_dates {
    my ($loc, %opts) = @_;
    my $start_date = $opts{start_date};
    my $end_date = $opts{end_date};
    # Actually around 2001-2, 
    # but no searchable data should be avail befoe 2000
    my $PMX_year_started = 2000;
    
    $start_date = '' if $start_date and $start_date eq $DEFAULT_DATE;
    $end_date = '' if $end_date and $end_date eq $DEFAULT_DATE;

    if ($opts{start_date_mandatory} and !$start_date and $end_date) {
        die $loc->("Please specify a start date\n");
    }
    return unless $start_date;

    if (!$end_date) {
        # Assume the current date/time
        $end_date = strftime(q{%Y-%m-%d %H:%M}, localtime);
    }
    my $start_range = $start_date;
    my $end_range   = $end_date;

    my ($s_year, $s_month, $s_day, $s_hour, $s_min, $s_sec);
    my ($e_year, $e_month, $e_day, $e_hour, $e_min, $e_sec);

    # Parse dates and validate formats

    my $start_date_format_valid = 0;
    my $end_date_format_valid = 0;

    $start_date_format_valid = 1
        if (($s_year, $s_month, $s_day, $s_hour, $s_min, $s_sec)
            = parse_date($start_range, 0));

    $end_date_format_valid = 1
        if (($e_year, $e_month, $e_day, $e_hour, $e_min, $e_sec)
            = parse_date($end_range, 1));

    if (!$start_date_format_valid && !$end_date_format_valid) {
        die PureMessage::WebUI::ValidationException->new(
                    error_msg  => $loc->("Invalid start/end date formats: [_1]/[_2]\n", $start_range, $end_range),
                    error_type => 'date_range',
                );
    }
    elsif (!$start_date_format_valid) {
        die PureMessage::WebUI::ValidationException->new(
                    error_msg  => $loc->("Invalid start date format: [_1]\n", $start_range),
                    error_type => 'date_range',
                );
    }
    elsif (!$end_date_format_valid) {
        die PureMessage::WebUI::ValidationException->new(
                    error_msg  => $loc->("Invalid end date format: [_1]\n", $end_range),
                    error_type => 'date_range',
                );
    }

    # Date validation (use the side effect from mktime)

    my $start_date_valid = 0;
    my $end_date_valid = 0;

    $start_date_valid = 1 if (check_date($s_year, $s_month, $s_day));
    $end_date_valid = 1 if (check_date($e_year, $e_month, $e_day));

    if (!$start_date_valid && !$end_date_valid) {
        die PureMessage::WebUI::ValidationException->new(
                    error_msg  => $loc->("Invalid start/end dates: [_1]/[_2]\n", $start_range, $end_range),
                    error_type => 'date_range',
                );
    }
    elsif (!$start_date_valid) {
        die PureMessage::WebUI::ValidationException->new(
                    error_msg  => $loc->("Invalid start date: [_1]\n", $start_range),
                    error_type => 'date_range',
                );
    }
    elsif (!$end_date_valid) {
        die PureMessage::WebUI::ValidationException->new(
                    error_msg  => $loc->("Invalid end date: [_1]\n", $end_range),
                    error_type => 'date_range',
                );
    }

    # check for valid hours and minutes
    for my $h ($s_hour, $e_hour) {
        unless ($h >= 0 and $h < 24) {
            die PureMessage::WebUI::ValidationException->new(
                        error_msg  => $loc->("Hour '[_1]' out of range 0..23\n", $h),
                        error_type => 'date_range',
                    );
        }
    }
    for my $m ($s_min, $e_min) {
        unless ($m >= 0 and $m < 60) {
            die PureMessage::WebUI::ValidationException->new(
                        error_msg  => $loc->("Minute '[_1]' out of range 0..59\n", $m),
                        error_type => 'date_range',
                    );
        }
    }

    my $s_epoch;
    eval {
        $s_epoch = Time::Local::timelocal(0, $s_min, $s_hour, $s_day,
                                         $s_month - 1, $s_year - 1900);
    };
    die PureMessage::WebUI::ValidationException->new(
                error_msg  => $loc->("Invalid start date: [_1]-[_2]-[_3]\n", $s_year, $s_month, $s_day),
                error_type => 'date_range',
            ) if $@ || $s_year < $PMX_year_started;

    my $e_epoch;
    eval {
        $e_epoch = Time::Local::timelocal(0, $e_min, $e_hour, $e_day,
                                         $e_month - 1, $e_year - 1900);
    };
    die PureMessage::WebUI::ValidationException->new(
                error_msg  => $loc->("Invalid end date: [_1]-[_2]-[_3]\n", $e_year, $e_month, $e_day),
                error_type => 'date_range',
            ) if $@ || $e_year < $PMX_year_started;

    die PureMessage::WebUI::ValidationException->new(
                error_msg  => $loc->("Invalid date range specified.\n"),
                error_type => 'date_range',
            ) if $s_epoch > $e_epoch;

    # Setup report options
    my $span_mins = ($e_epoch - $s_epoch) / 60;
    return ($start_range, $span_mins, $end_date);
}

sub parse_date {
    my ($date, $is_end_date) = @_;
    my ($year, $month, $day, $hour, $min);
    my $default_hour = $is_end_date ? 23 : 0;
    my $default_min = $is_end_date ? 59 : 0;
    # Date format: YYYY-MM-DD HH:MM
    if ($date =~ m{^\s*(\d{4})[-/](\d{1,2})[-/](\d{1,2}) # YYYY-MM-DD
                   (?:\s+(\d{1,2}):(\d{2}))?         # HH:MM
                   \s*$}x) {
        ($year, $month, $day, $hour, $min) = ($1, $2, $3, 
                                              $4 || $default_hour, 
                                              $5 || $default_min);
        my $sec = $is_end_date ? 59 : 0;
        return ($year, $month, $day, $hour, $min, $sec);
    }
    return;
}

# Replacement for the Template::Toolkit truncate filter which truncates on
# byte boundaries instead of character boundaries.  This results in multi-byte
# characters being truncated mid-character and invalid utf8 strings.
sub utf8_truncate_filter {
    my ($context, $length) = @_;    
    return sub {
        my $text = shift;
        my $new_length = $length - 3;  # for the ...
        if (is_valid_utf8($text)) {
            # ensure new length is on a character boundary
            $new_length = utf8_length($text, $new_length);
        }
        return substr($text, 0, $new_length) . '...';
    };
}

1;
